import random
import numpy as np
import torch
from torch.utils.data import DataLoader
from lightwood.model.helpers.ranger import Ranger
from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.gym import Gym
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.model.helpers.default_net import DefaultNet


class CategoricalAutoEncoder(BaseEncoder):
    def __init__(self, stop_after: int = 3600, is_target=False, max_encoded_length=100):
        super().__init__(is_target)
        self._prepared = False
        self.name = 'Categorical Autoencoder'
        self.net = None
        self.encoder = None
        self.decoder = None
        self.onehot_encoder = OneHotEncoder(is_target=self.is_target)
        self.desired_error = 0.01
        self.use_autoencoder = None
        self.stop_after = stop_after
        # @TODO stop using instead of ONEHOT !!!@!
        self.is_nn_encoder = True
        self.output_size = None
        if self.is_target:
            self.max_encoded_length = None
        else:
            self.max_encoded_length = max_encoded_length

    def _train_callback(self, error, real_buff, predicted_buff):
        log.info(f'{self.name} reached a loss of {error} while training !')

    def _encoder_targets(self, data):
        oh_encoded_categories = self.onehot_encoder.encode(data)
        target = oh_encoded_categories.cpu().numpy()
        target_indexes = np.where(target > 0)[1]
        targets_c = torch.LongTensor(target_indexes)
        labels = targets_c.to(self.net.device)
        return labels

    def prepare(self, priming_data):
        random.seed(len(priming_data))

        if self._prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.onehot_encoder.prepare(priming_data)

        input_len = self.onehot_encoder._lang.n_words
        self.use_autoencoder = self.max_encoded_length is not None and input_len > self.max_encoded_length

        if self.use_autoencoder:
            log.info('Preparing a categorical autoencoder, this might take a while')

            embeddings_layer_len = self.max_encoded_length

            self.net = DefaultNet(shape=[
                                  input_len, embeddings_layer_len, input_len])

            criterion = torch.nn.CrossEntropyLoss()
            optimizer = Ranger(self.net.parameters())

            gym = Gym(model=self.net, optimizer=optimizer, scheduler=None, loss_criterion=criterion,
                      device=self.net.device, name=self.name, input_encoder=self.onehot_encoder.encode,
                      output_encoder=self._encoder_targets)

            batch_size = min(200, int(len(priming_data) / 50))

            priming_data_str = [str(x) for x in priming_data]
            train_data_loader = DataLoader(
                list(zip(priming_data_str, priming_data_str)),
                batch_size=batch_size, shuffle=True)

            test_data_loader = None

            best_model, error, training_time = gym.fit(train_data_loader,
                                                       test_data_loader,
                                                       desired_error=self.desired_error,
                                                       max_time=self.stop_after,
                                                       callback=self._train_callback,
                                                       eval_every_x_epochs=1,
                                                       max_unimproving_models=5)

            self.net = best_model.to(self.net.device)

            modules = [module for module in self.net.modules() if type(
                module) != torch.nn.Sequential and type(module) != DefaultNet]
            self.encoder = torch.nn.Sequential(*modules[0:2]).eval()
            self.decoder = torch.nn.Sequential(*modules[2:3]).eval()
            log.info('Categorical autoencoder ready')

        self.output_size = self.onehot_encoder._lang.n_words
        self._prepared = True

    def encode(self, column_data):
        oh_encoded_tensor = self.onehot_encoder.encode(column_data)
        if not self.use_autoencoder:
            return oh_encoded_tensor
        else:
            with torch.no_grad():
                oh_encoded_tensor = oh_encoded_tensor.to(self.net.device)
                embeddings = self.encoder(oh_encoded_tensor)
                return embeddings.to('cpu')

    def decode(self, encoded_data):
        if not self.use_autoencoder:
            return self.onehot_encoder.decode(encoded_data)
        else:
            with torch.no_grad():
                encoded_data = encoded_data.to(self.net.device)
                oh_encoded_tensor = self.decoder(encoded_data)
                oh_encoded_tensor = oh_encoded_tensor.to('cpu')
                return self.onehot_encoder.decode(oh_encoded_tensor)



from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.binary import BinaryEncoder
from lightwood.encoder.time_series.rnn import TimeSeriesEncoder
from lightwood.encoder.time_series.plain import TimeSeriesPlainEncoder


__all__ = ['BinaryEncoder', 'OneHotEncoder', 'TimeSeriesEncoder', 'TimeSeriesPlainEncoder']



from lightwood import Predictor
from lightwood.constants.lightwood import ColumnDataTypes
from collections import Counter
import numpy as np
import pickle
import os


class LightwoodEnsemble:
    def __init__(self, predictors=None, load_from_path=None):
        self.path_list = None
        if load_from_path is not None:
            with open(os.path.join(load_from_path, 'lightwood_data'), 'rb') as pickle_in:
                obj = pickle.load(pickle_in)
                self.path = load_from_path
                self.path_list = obj.path_list
                self.ensemble = [Predictor(load_from_path=path) for path in self.path_list]
        elif isinstance(predictors, Predictor):
            self.ensemble = [predictors]
        elif isinstance(predictors, list):
            self.ensemble = predictors

    def append(self, predictor):
        if isinstance(self.ensemble, list):
            self.ensemble.append(predictor)
        else:
            self.ensemble = [predictor]

    def __iter__(self):
        yield self.ensemble

    def predict(self, when_data):
        predictions = [p.predict(when_data=when_data) for p in self.ensemble]
        formatted_predictions = {}
        for target in self.ensemble[0].config['output_features']:
            target_name = target['name']
            formatted_predictions[target_name] = {}
            pred_arr = np.array([p[target_name]['predictions'] for p in predictions])
            if target['type'] == ColumnDataTypes.NUMERIC:
                final_preds = np.mean(pred_arr, axis=0).tolist()
            elif target['type'] == ColumnDataTypes.CATEGORICAL:
                final_preds = [max(Counter(pred_arr[:, idx])) for idx in range(pred_arr.shape[1])]

                # @TODO: implement class distribution for ensembles
                # NOTE: label set *could* grow when adding predictors, which complicates belief score computation
                formatted_predictions[target_name]['class_distribution'] = np.ones(shape=(len(final_preds), 1))
            else:
                raise Exception('Only numeric and categorical datatypes are supported for ensembles')

            formatted_predictions[target_name]['predictions'] = final_preds

        return formatted_predictions

    def save(self, path_to):
        # TODO: potentially save predictors inside ensemble pickle, though there's the issue of nonpersistent stuff with torch.save() # noqa
        path_list = []
        for i, model in enumerate(self.ensemble):
            path = os.path.join(path_to, f'lightwood_predictor_{i}')
            path_list.append(path)
            model.save(path_to=path)

        self.path_list = path_list

        # TODO: in the future, save preds inside this data struct
        self.ensemble = None  # we deref predictors for now
        with open(os.path.join(path_to, 'lightwood_data'), 'wb') as file:
            pickle.dump(self, file, pickle.HIGHEST_PROTOCOL)



from typing import List
import pandas as pd
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs


def encode(encoders: List[BaseEncoder], folds: List[pd.DataFrame], target: str) -> List[EncodedDs]:
    if isinstance(folds, pd.DataFrame):
        folds = [folds]

    encoded_ds_arr: List[EncodedDs] = []
    for fold in folds:
        encoded_ds_arr.append(EncodedDs(encoders, fold, target))
    return encoded_ds_arr



from typing import Dict
from lightwood.helpers.templating import call, inline_dict, align
import autopep8
from lightwood.api import dtype
import numpy as np
from lightwood.api.types import (JsonAI, TypeInformation, StatisticalAnalysis, Feature,
                                 Output, ProblemDefinition)


trainable_encoders = ('PretrainedLangEncoder', 'CategoricalAutoEncoder', 'TimeSeriesEncoder', 'TimeSeriesPlainEncoder')
ts_encoders = ('TimeSeriesEncoder', 'TimeSeriesPlainEncoder', 'TsNumericEncoder')


def lookup_encoder(
        col_dtype: str, col_name: str, is_target: bool, problem_defintion: ProblemDefinition,
        is_target_predicting_encoder: bool):
    tss = problem_defintion.timeseries_settings
    encoder_lookup = {
        dtype.integer: 'Integer.NumericEncoder',
        dtype.float: 'Float.NumericEncoder',
        dtype.binary: 'Binary.BinaryEncoder',
        dtype.categorical: 'Categorical.CategoricalAutoEncoder',
        dtype.tags: 'Tags.MultiHotEncoder',
        dtype.date: 'Date.DatetimeEncoder',
        dtype.datetime: 'Datetime.DatetimeEncoder',
        dtype.image: 'Image.Img2VecEncoder',
        dtype.rich_text: 'Rich_Text.PretrainedLangEncoder',
        dtype.short_text: 'Short_Text.CategoricalAutoEncoder',
        dtype.array: 'Array.TimeSeriesEncoder',
        dtype.quantity: 'Quantity.NumericEncoder',
    }

    target_encoder_lookup_override = {
        dtype.rich_text: 'Rich_Text.VocabularyEncoder',
        dtype.categorical: 'Categorical.OneHotEncoder'
    }

    encoder_dict = {
        'module': encoder_lookup[col_dtype],
        'args': {}
    }

    if is_target:
        encoder_dict['args'] = {'is_target': 'True'}
        if col_dtype in target_encoder_lookup_override:
            encoder_dict['module'] = target_encoder_lookup_override[col_dtype]
        if col_dtype in (dtype.categorical, dtype.binary):
            if problem_defintion.unbias_target:
                encoder_dict['args'] = {'target_class_distribution': '$statistical_analysis.target_class_distribution'}

    if tss.is_timeseries:
        gby = tss.group_by if tss.group_by is not None else []
        if col_name in tss.order_by + tss.historical_columns:
            encoder_dict['module'] = col_dtype.capitalize() + '.TimeSeriesEncoder'
            encoder_dict['args']['original_type'] = f'"{col_dtype}"'
            encoder_dict['args']['target'] = "self.target"
            encoder_dict['args']['grouped_by'] = f"{gby}"
        if is_target:
            if col_dtype in [dtype.integer]:
                encoder_dict['args']['grouped_by'] = f"{gby}"
                encoder_dict['module'] = 'Integer.TsNumericEncoder'
            if col_dtype in [dtype.float]:
                encoder_dict['args']['grouped_by'] = f"{gby}"
                encoder_dict['module'] = 'Float.TsNumericEncoder'
            if tss.nr_predictions > 1:
                encoder_dict['args']['grouped_by'] = f"{gby}"
                encoder_dict['args']['timesteps'] = f"{tss.nr_predictions}"
                encoder_dict['module'] = 'Array.TsArrayNumericEncoder'
        if '__mdb_ts_previous' in col_name:
            encoder_dict['module'] = col_dtype.capitalize() + '.TimeSeriesPlainEncoder'
            encoder_dict['args']['original_type'] = f'"{tss.target_type}"'
            encoder_dict['args']['window'] = f'{tss.window}'

    # Set arguments for the encoder
    if encoder_dict['module'] == 'Rich_Text.PretrainedLangEncoder' and not is_target:
        encoder_dict['args']['output_type'] = '$dtype_dict[$target]'

    for encoder_name in trainable_encoders:
        if encoder_name == encoder_dict['module'].split('.')[1]:
            encoder_dict['args']['stop_after'] = '$problem_definition.seconds_per_encoder'

    if is_target_predicting_encoder:
        encoder_dict['args']['embed_mode'] = 'False'

    return encoder_dict


def generate_json_ai(type_information: TypeInformation, statistical_analysis: StatisticalAnalysis,
                     problem_definition: ProblemDefinition) -> JsonAI:
    target = problem_definition.target
    input_cols = []
    for col_name, col_dtype in type_information.dtypes.items():
        if col_name not in type_information.identifiers and col_dtype not in (
                dtype.invalid, dtype.empty) and col_name != target:
            input_cols.append(col_name)

    is_target_predicting_encoder = False
    # Single text column classification
    if len(input_cols) == 1 and type_information.dtypes[
            input_cols[0]] in (
            dtype.rich_text) and type_information.dtypes[target] in (
            dtype.categorical, dtype.binary):
        is_target_predicting_encoder = True

    if is_target_predicting_encoder:
        models = [{
            'module': 'Unit',
            'args': {
                'target_encoder': '$encoders[self.target]',
                'stop_after': '$problem_definition.seconds_per_model'
            }
        }]
    else:
        models = [{
            'module': 'Neural',
            'args': {
                'fit_on_dev': True,
                'stop_after': '$problem_definition.seconds_per_model',
                'search_hyperparameters': True
            }

        }]

        if not problem_definition.timeseries_settings.is_timeseries or \
                problem_definition.timeseries_settings.nr_predictions <= 1:
            models.extend([{
                'module': 'LightGBM',
                'args': {
                    'stop_after': '$problem_definition.seconds_per_model',
                    'fit_on_dev': True
                }
            },
                {
                    'module': 'Regression',
                    'args': {
                        'stop_after': '$problem_definition.seconds_per_model',
                    }
            }
            ])
        elif problem_definition.timeseries_settings.nr_predictions > 1:
            models.extend([{
                'module': 'LightGBMArray',
                'args': {
                    'fit_on_dev': True,
                    'stop_after': '$problem_definition.seconds_per_model',
                    'n_ts_predictions': '$problem_definition.timeseries_settings.nr_predictions'
                }
            },
                {
                'module': 'SkTime',
                'args': {
                    'stop_after': '$problem_definition.seconds_per_model',
                    'n_ts_predictions': '$problem_definition.timeseries_settings.nr_predictions',
                },
            }
            ])

    outputs = {target: Output(
        data_dtype=type_information.dtypes[target],
        encoder=None,
        models=models,
        ensemble={
            'module': 'BestOf',
            'args': {
                'accuracy_functions': '$accuracy_functions',
            }
        }
    )}

    if (problem_definition.timeseries_settings.is_timeseries and
            problem_definition.timeseries_settings.nr_predictions > 1):
        list(outputs.values())[0].data_dtype = dtype.array

    list(
        outputs.values())[0].encoder = lookup_encoder(
        type_information.dtypes[target],
        target, True, problem_definition, False)

    features: Dict[str, Feature] = {}
    for col_name in input_cols:
        col_dtype = type_information.dtypes[col_name]
        dependency = []
        encoder = lookup_encoder(col_dtype, col_name, False, problem_definition, is_target_predicting_encoder)

        for encoder_name in ts_encoders:
            if problem_definition.timeseries_settings.is_timeseries and encoder_name == encoder['module'].split('.')[1]:
                if problem_definition.timeseries_settings.group_by is not None:
                    for group in problem_definition.timeseries_settings.group_by:
                        dependency.append(group)

                if problem_definition.timeseries_settings.use_previous_target:
                    dependency.append(f'__mdb_ts_previous_{target}')

        if len(dependency) > 0:
            feature = Feature(
                encoder=encoder,
                dependency=dependency
            )
        else:
            feature = Feature(
                encoder=encoder
            )
        features[col_name] = feature

    # Decide on the accuracy functions to use
    if list(outputs.values())[0].data_dtype in [dtype.integer, dtype.float]:
        accuracy_functions = ['r2_score']
    elif list(outputs.values())[0].data_dtype == dtype.categorical:
        accuracy_functions = ['balanced_accuracy_score']
    elif list(outputs.values())[0].data_dtype == dtype.tags:
        accuracy_functions = ['balanced_accuracy_score']
    elif list(outputs.values())[0].data_dtype == dtype.array:
        accuracy_functions = ['evaluate_array_accuracy']
    else:
        accuracy_functions = ['accuracy_score']

    if problem_definition.time_aim is None and (
            problem_definition.seconds_per_model is None or problem_definition.seconds_per_encoder is None):
        problem_definition.time_aim = 1000 + np.log(
            statistical_analysis.nr_rows / 10 + 1) * np.sum(
            [4
             if x in [dtype.rich_text, dtype.short_text, dtype.array, dtype.video, dtype.audio, dtype.image] else 1
             for x in type_information.dtypes.values()]) * 200

    if problem_definition.time_aim is not None:
        nr_trainable_encoders = len([x for x in features.values() if x.encoder['module'].split('.')[1]
                                    in trainable_encoders])
        nr_models = len(list(outputs.values())[0].models)
        encoder_time_budget_pct = max(3.3 / 5, 1.5 + np.log(nr_trainable_encoders + 1) / 5)

        if nr_trainable_encoders == 0:
            problem_definition.seconds_per_encoder = 0
        else:
            problem_definition.seconds_per_encoder = int(
                problem_definition.time_aim * (encoder_time_budget_pct / nr_trainable_encoders))
        problem_definition.seconds_per_model = int(
            problem_definition.time_aim * ((1 / encoder_time_budget_pct) / nr_models))

    return JsonAI(
        cleaner=None,
        splitter=None,
        analyzer=None,
        explainer=None,
        features=features,
        outputs=outputs,
        imports=None,
        problem_definition=problem_definition,
        identifiers=type_information.identifiers,
        timeseries_transformer=None,
        timeseries_analyzer=None,
        accuracy_functions=accuracy_functions
    )


def add_implicit_values(json_ai: JsonAI) -> JsonAI:
    problem_definition = json_ai.problem_definition
    imports = [
        'from lightwood.model import Neural', 'from lightwood.model import LightGBM',
        'from lightwood.model import LightGBMArray', 'from lightwood.model import SkTime',
        'from lightwood.model import Unit', 'from lightwood.model import Regression',
        'from lightwood.ensemble import BestOf', 'from lightwood.data import cleaner',
        'from lightwood.data import transform_timeseries, timeseries_analyzer', 'from lightwood.data import splitter',
        'from lightwood.analysis import model_analyzer, explain',
        'from sklearn.metrics import r2_score, balanced_accuracy_score, accuracy_score', 'import pandas as pd',
        'from lightwood.helpers.seed import seed', 'from lightwood.helpers.log import log', 'import lightwood',
        'from lightwood.api import *', 'from lightwood.model import BaseModel',
        'from lightwood.encoder import BaseEncoder, __ts_encoders__',
        'from lightwood.encoder import Array, Binary, Categorical, Date, Datetime, Float, Image, Integer, Quantity, Rich_Text, Short_Text, Tags', # noqa
        'from lightwood.ensemble import BaseEnsemble', 'from typing import Dict, List',
        'from lightwood.helpers.parallelism import mut_method_call',
        'from lightwood.data.encoded_ds import ConcatedEncodedDs', 'from lightwood import ProblemDefinition']

    if json_ai.imports is None:
        json_ai.imports = imports
    else:
        json_ai.imports.extend(imports)

    for feature in [list(json_ai.outputs.values())[0], *json_ai.features.values()]:
        encoder_import = feature.encoder['module']
        if '.' in encoder_import:
            continue
        imports.append(f'from lightwood.encoder import {encoder_import}')

    if problem_definition.timeseries_settings.use_previous_target:
        imports.append('from lightwood.encoder import TimeSeriesPlainEncoder')

    # Add implicit arguments
    # @TODO: Consider removing once we have a proper editor in studio
    models = json_ai.outputs[json_ai.problem_definition.target].models
    for i in range(len(models)):
        if models[i]['module'] == 'Unit':
            pass
        elif models[i]['module'] == 'Neural':
            models[i]['args']['target_encoder'] = models[i]['args'].get('target_encoder', '$encoders[self.target]')
            models[i]['args']['target'] = models[i]['args'].get('target', '$target')
            models[i]['args']['dtype_dict'] = models[i]['args'].get('dtype_dict', '$dtype_dict')
            models[i]['args']['input_cols'] = models[i]['args'].get('input_cols', '$input_cols')
            models[i]['args']['timeseries_settings'] = models[i]['args'].get(
                'timeseries_settings', '$problem_definition.timeseries_settings')
            models[i]['args']['net'] = models[i]['args'].get(
                'net', '"DefaultNet"' if not problem_definition.timeseries_settings.is_timeseries else '"ArNet"')

        elif models[i]['module'] == 'LightGBM':
            models[i]['args']['target'] = models[i]['args'].get('target', '$target')
            models[i]['args']['dtype_dict'] = models[i]['args'].get('dtype_dict', '$dtype_dict')
            models[i]['args']['input_cols'] = models[i]['args'].get('input_cols', '$input_cols')
        elif models[i]['module'] == 'Regression':
            models[i]['args']['target'] = models[i]['args'].get('target', '$target')
            models[i]['args']['dtype_dict'] = models[i]['args'].get('dtype_dict', '$dtype_dict')
            models[i]['args']['target_encoder'] = models[i]['args'].get('target_encoder', '$encoders[self.target]')
        elif models[i]['module'] == 'LightGBMArray':
            models[i]['args']['target'] = models[i]['args'].get('target', '$target')
            models[i]['args']['dtype_dict'] = models[i]['args'].get('dtype_dict', '$dtype_dict')
            models[i]['args']['input_cols'] = models[i]['args'].get('input_cols', '$input_cols')
        elif models[i]['module'] == 'SkTime':
            models[i]['args']['target'] = models[i]['args'].get('target', '$target')
            models[i]['args']['dtype_dict'] = models[i]['args'].get('dtype_dict', '$dtype_dict')
            models[i]['args']['ts_analysis'] = models[i]['args'].get('ts_analysis', '$ts_analysis')

    ensemble = json_ai.outputs[json_ai.problem_definition.target].ensemble
    ensemble['args']['target'] = ensemble['args'].get('target', '$target')
    ensemble['args']['data'] = ensemble['args'].get('data', 'test_data')
    ensemble['args']['models'] = ensemble['args'].get('models', '$models')

    for name in json_ai.features:
        if json_ai.features[name].dependency is None:
            json_ai.features[name].dependency = []
        if json_ai.features[name].data_dtype is None:
            json_ai.features[name].data_dtype = json_ai.features[name].encoder['module'].split('.')[0].lower()

    # Add implicit phases
    # @TODO: Consider removing once we have a proper editor in studio
    if json_ai.cleaner is None:
        json_ai.cleaner = {
            'module': 'cleaner',
            'args': {
                'pct_invalid': '$problem_definition.pct_invalid',
                'ignore_features': '$problem_definition.ignore_features',
                'identifiers': '$identifiers',
                'data': 'data',
                'dtype_dict': '$dtype_dict',
                'target': '$target',
                'mode': '$mode',
                'timeseries_settings': '$problem_definition.timeseries_settings'
            }
        }

    if json_ai.splitter is None:
        json_ai.splitter = {
            'module': 'splitter',
            'args': {
                'tss': '$problem_definition.timeseries_settings',
                'data': 'data',
                'k': 'nfolds'
            }
        }
    if json_ai.analyzer is None:
        json_ai.analyzer = {
            'module': 'model_analyzer',
            'args': {
                'stats_info': '$statistical_analysis',
                'ts_cfg': '$problem_definition.timeseries_settings',
                'accuracy_functions': '$accuracy_functions',
                'predictor': '$ensemble',
                'data': 'test_data',
                'train_data': 'train_data',
                'target': '$target',
                'disable_column_importance': 'False',
                'dtype_dict': '$dtype_dict',
                'fixed_significance': None,
                'confidence_normalizer': False,
                'positive_domain': False,
            }
        }

    if json_ai.explainer is None:
        json_ai.explainer = {
            'module': 'explain',
            'args': {
                'timeseries_settings': '$problem_definition.timeseries_settings',
                'positive_domain': '$problem_definition.positive_domain',
                'fixed_confidence': '$problem_definition.fixed_confidence',
                'anomaly_detection': '$problem_definition.anomaly_detection',
                'anomaly_error_rate': '$problem_definition.anomaly_error_rate',
                'anomaly_cooldown': '$problem_definition.anomaly_cooldown',
                'data': 'data',
                'encoded_data': 'encoded_data',
                'predictions': 'df',
                'analysis': '$runtime_analyzer',
                'ts_analysis': '$ts_analysis' if problem_definition.timeseries_settings.is_timeseries else None,
                'target_name': '$target',
                'target_dtype': '$dtype_dict[self.target]',
            }
        }

    if problem_definition.timeseries_settings.is_timeseries:
        if json_ai.timeseries_transformer is None:
            json_ai.timeseries_transformer = {
                'module': 'transform_timeseries',
                'args': {
                    'timeseries_settings': '$problem_definition.timeseries_settings',
                    'data': 'data',
                    'dtype_dict': '$dtype_dict',
                    'target': '$target',
                    'mode': '$mode'
                }
            }

        if json_ai.timeseries_analyzer is None:
            json_ai.timeseries_analyzer = {
                'module': 'timeseries_analyzer',
                'args': {
                    'timeseries_settings': '$problem_definition.timeseries_settings',
                    'data': 'data',
                    'dtype_dict': '$dtype_dict',
                    'target': '$target'
                }
            }

    return json_ai


def code_from_json_ai(json_ai: JsonAI) -> str:
    json_ai = add_implicit_values(json_ai)

    encoder_dict = {json_ai.problem_definition.target: call(list(json_ai.outputs.values())[0].encoder, json_ai)}
    dependency_dict = {}
    dtype_dict = {json_ai.problem_definition.target: f"""'{list(json_ai.outputs.values())[0].data_dtype}'"""}

    for col_name, feature in json_ai.features.items():
        if col_name not in json_ai.problem_definition.ignore_features:
            encoder_dict[col_name] = call(feature.encoder, json_ai)
            dependency_dict[col_name] = feature.dependency
            dtype_dict[col_name] = f"""'{feature.data_dtype}'"""

    # @TODO: Move into json-ai creation function (I think? Maybe? Let's discuss)
    if json_ai.problem_definition.timeseries_settings.use_previous_target:
        col_name = f'__mdb_ts_previous_{json_ai.problem_definition.target}'
        json_ai.problem_definition.timeseries_settings.target_type = list(json_ai.outputs.values())[0].data_dtype
        encoder_dict[col_name] = call(lookup_encoder(list(json_ai.outputs.values())[0].data_dtype,
                                                     col_name,
                                                     False,
                                                     json_ai.problem_definition,
                                                     False,
                                                     ),
                                      json_ai)
        dependency_dict[col_name] = []
        dtype_dict[col_name] = f"""'{list(json_ai.outputs.values())[0].data_dtype}'"""
        json_ai.features[col_name] = Feature(
            encoder=encoder_dict[col_name]
        )

    ignored_cols = json_ai.problem_definition.ignore_features
    input_cols = ','.join([f"""'{name}'""" for name in json_ai.features if name not in ignored_cols])

    ts_transform_code = ''
    ts_analyze_code = ''
    ts_encoder_code = ''
    if json_ai.timeseries_transformer is not None:
        ts_transform_code = f"""
log.info('Transforming timeseries data')
data = {call(json_ai.timeseries_transformer, json_ai)}
"""
        ts_analyze_code = f"""
self.ts_analysis = {call(json_ai.timeseries_analyzer, json_ai)}
"""

    if json_ai.timeseries_analyzer is not None:
        ts_encoder_code = """
if type(encoder) in __ts_encoders__:
    kwargs['ts_analysis'] = self.ts_analysis
"""

    if json_ai.problem_definition.timeseries_settings.is_timeseries:
        ts_target_code = """
if encoder.is_target:
    encoder.normalizers = self.ts_analysis['target_normalizers']
    encoder.group_combinations = self.ts_analysis['group_combinations']
"""
    else:
        ts_target_code = ''

    dataprep_body = f"""
# The type of each column
self.problem_definition = ProblemDefinition.from_dict({json_ai.problem_definition.to_dict()})
self.accuracy_functions = {json_ai.accuracy_functions}
self.identifiers = {json_ai.identifiers}
self.dtype_dict = {inline_dict(dtype_dict)}
self.statistical_analysis = lightwood.data.statistical_analysis(data, self.dtype_dict, {json_ai.identifiers},
                                                                self.problem_definition)
self.mode = 'train'
# How columns are encoded
self.encoders = {inline_dict(encoder_dict)}
# Which column depends on which
self.dependencies = {inline_dict(dependency_dict)}
#
self.input_cols = [{input_cols}]

log.info('Cleaning the data')
data = {call(json_ai.cleaner, json_ai)}

{ts_transform_code}
{ts_analyze_code}

nfolds = {json_ai.problem_definition.nfolds}
log.info(f'Splitting the data into {{nfolds}} folds')
folds = {call(json_ai.splitter, json_ai)}

log.info('Preparing the encoders')

encoder_preping_dict = {{}}
enc_preping_data = pd.concat(folds[0:nfolds-1])
for col_name, encoder in self.encoders.items():
    if not encoder.is_nn_encoder:
        encoder_preping_dict[col_name] = [encoder, enc_preping_data[col_name], 'prepare']
        log.info(f'Encoder preping dict length of: {{len(encoder_preping_dict)}}')

parallel_preped_encoders = mut_method_call(encoder_preping_dict)
for col_name, encoder in parallel_preped_encoders.items():
    self.encoders[col_name] = encoder

if self.target not in parallel_preped_encoders:
    self.encoders[self.target].prepare(enc_preping_data[self.target])

for col_name, encoder in self.encoders.items():
    if encoder.is_nn_encoder:
        priming_data = pd.concat(folds[0:nfolds-1])
        kwargs = {{}}
        if self.dependencies[col_name]:
            kwargs['dependency_data'] = {{}}
            for col in self.dependencies[col_name]:
                kwargs['dependency_data'][col] = {{
                    'original_type': self.dtype_dict[col],
                    'data': priming_data[col]
                }}
            {align(ts_encoder_code, 3)}

        # This assumes target  encoders are also prepared in parallel, might not be true
        if hasattr(encoder, 'uses_target'):
            kwargs['encoded_target_values'] = parallel_preped_encoders[self.target].encode(priming_data[self.target])

        encoder.prepare(priming_data[col_name], **kwargs)

    {align(ts_target_code, 1)}
"""
    dataprep_body = align(dataprep_body, 2)

    learn_body = f"""
log.info('Featurizing the data')
encoded_ds_arr = lightwood.encode(self.encoders, folds, self.target)
train_data = encoded_ds_arr[0:int(nfolds*0.9)]
test_data = encoded_ds_arr[int(nfolds*0.9):]

log.info('Training the models')
self.models = [{', '.join([call(x, json_ai) for x in list(json_ai.outputs.values())[0].models])}]
trained_models = []
for model in self.models:
    try:
        model.fit(train_data)
        trained_models.append(model)
    except Exception as e:
        log.warning(f'Exception: {{e}} when training model: {{model}}')
        if {json_ai.problem_definition.strict_mode} and model.stable:
            raise e

self.models = trained_models

log.info('Ensembling the model')
self.ensemble = {call(list(json_ai.outputs.values())[0].ensemble, json_ai)}
self.supports_proba = self.ensemble.supports_proba

log.info('Analyzing the ensemble')
self.model_analysis, self.runtime_analyzer = {call(json_ai.analyzer, json_ai)}

# Partially fit the model on the reamining of the data, data is precious, we mustn't loss one bit
for model in self.models:
    if {json_ai.problem_definition.fit_on_validation}:
        model.partial_fit(test_data, train_data)
"""
    learn_body = align(learn_body, 2)

    predict_common_body = f"""
self.mode = 'predict'
log.info('Cleaning the data')
data = {call(json_ai.cleaner, json_ai)}

{ts_transform_code}

encoded_ds = lightwood.encode(self.encoders, data, self.target)[0]
encoded_data = encoded_ds.get_encoded_data(include_target=False)
"""
    predict_common_body = align(predict_common_body, 2)

    predict_body = f"""
df = self.ensemble(encoded_ds)
insights = {call(json_ai.explainer, json_ai)}
return insights
"""
    predict_body = align(predict_body, 2)

    predict_proba_body = f"""
df = self.ensemble(encoded_ds, predict_proba=True)
insights = {call(json_ai.explainer, json_ai)}
return insights
"""
    predict_proba_body = align(predict_proba_body, 2)

    imports = '\n'.join(json_ai.imports)
    predictor_code = f"""
{imports}
from lightwood.api import PredictorInterface


class Predictor(PredictorInterface):
    target: str
    models: List[BaseModel]
    encoders: Dict[str, BaseEncoder]
    ensemble: BaseEnsemble
    mode: str

    def __init__(self):
        seed()
        self.target = '{json_ai.problem_definition.target}'
        self.mode = 'innactive'

    def learn(self, data: pd.DataFrame) -> None:
{dataprep_body}
{learn_body}

    def predict(self, data: pd.DataFrame) -> pd.DataFrame:
{predict_common_body}
{predict_body}


    def predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
{predict_common_body}
{predict_proba_body}
"""

    if len(predictor_code) < 5000:
        predictor_code = autopep8.fix_code(predictor_code)

    return predictor_code


def validate_json_ai(json_ai: JsonAI) -> bool:
    from lightwood.api.high_level import predictor_from_code, code_from_json_ai
    try:
        predictor_from_code(code_from_json_ai(json_ai))
        return True
    except Exception:
        return False



from typing import Dict, List

import torch
import numpy as np
import pandas as pd
from copy import deepcopy
from itertools import product
from sklearn.preprocessing import OneHotEncoder

from lightwood.api import dtype
from lightwood.api.types import ModelAnalysis, StatisticalAnalysis, TimeseriesSettings
from lightwood.data.encoded_ds import ConcatedEncodedDs, EncodedDs
from lightwood.helpers.general import evaluate_accuracy
from lightwood.ensemble import BaseEnsemble
from lightwood.encoder.text.pretrained import PretrainedLangEncoder

from lightwood.analysis.acc_stats import AccStats
from lightwood.analysis.nc.norm import Normalizer
from lightwood.analysis.nc.nc import BoostedAbsErrorErrFunc
from lightwood.analysis.nc.util import clean_df, set_conf_range
from lightwood.analysis.nc.icp import IcpRegressor, IcpClassifier
from lightwood.analysis.nc.nc import RegressorNc, ClassifierNc, MarginErrFunc
from lightwood.analysis.nc.wrappers import ConformalClassifierAdapter, ConformalRegressorAdapter, t_softmax


"""
Pending:
 - [] simplify nonconformist custom implementation to deprecate wrappers
 - [] reimplement caching for faster analysis?
 - [] confidence for T+N <- active research question
"""


def model_analyzer(
    predictor: BaseEnsemble,
    data: List[EncodedDs],
    train_data: List[EncodedDs],
    stats_info: StatisticalAnalysis,
    target: str,
    ts_cfg: TimeseriesSettings,
    dtype_dict: Dict[str, str],
    disable_column_importance: bool,
    fixed_significance: float,
    positive_domain: bool,
    confidence_normalizer: bool,
    accuracy_functions
):
    """Analyses model on a validation fold to evaluate accuracy and confidence of future predictions"""

    data_type = dtype_dict[target]
    data_subtype = data_type

    is_numerical = data_type in [dtype.integer, dtype.float] or data_type in [dtype.array]
    is_classification = data_type in (dtype.categorical, dtype.binary)
    is_multi_ts = ts_cfg.is_timeseries and ts_cfg.nr_predictions > 1

    encoded_train_data = ConcatedEncodedDs(train_data)
    encoded_data = ConcatedEncodedDs(data)

    has_pretrained_text_enc = any([isinstance(enc, PretrainedLangEncoder)
                                   for enc in encoded_train_data.encoders.values()])
    disable_column_importance = disable_column_importance or ts_cfg.is_timeseries or has_pretrained_text_enc

    data = encoded_data.data_frame
    runtime_analyzer = {}
    predictions = {}
    input_cols = list([col for col in data.columns if col != target])
    normal_predictions = predictor(encoded_data) if not is_classification else predictor(
        encoded_data, predict_proba=True)
    normal_predictions = normal_predictions.set_index(data.index)

    # confidence estimation with inductive conformal predictors (ICPs)
    runtime_analyzer['icp'] = {'__mdb_active': False}

    fit_params = {'nr_preds': ts_cfg.nr_predictions or 0, 'columns_to_ignore': []}
    fit_params['columns_to_ignore'].extend([f'timestep_{i}' for i in range(1, fit_params['nr_preds'])])

    if is_classification:
        if predictor.supports_proba:
            all_cat_cols = [col for col in normal_predictions.columns if '__mdb_proba' in col]
            all_classes = np.array([col.replace('__mdb_proba_', '') for col in all_cat_cols])
        else:
            class_keys = sorted(encoded_data.encoders[target].rev_map.keys())
            all_classes = np.array([encoded_data.encoders[target].rev_map[idx] for idx in class_keys])

        if data_subtype != dtype.tags:
            enc = OneHotEncoder(sparse=False, handle_unknown='ignore')
            enc.fit(all_classes.reshape(-1, 1))
            runtime_analyzer['label_encoders'] = enc  # needed to repr cat labels inside nonconformist
        else:
            runtime_analyzer['label_encoders'] = None

        adapter = ConformalClassifierAdapter
        nc_function = MarginErrFunc()
        nc_class = ClassifierNc
        icp_class = IcpClassifier

    else:
        adapter = ConformalRegressorAdapter
        nc_function = BoostedAbsErrorErrFunc()
        nc_class = RegressorNc
        icp_class = IcpRegressor

    if is_numerical or (is_classification and data_subtype != dtype.tags):
        model = adapter(predictor)

        norm_params = {'target': target, 'dtype_dict': dtype_dict, 'predictor': predictor,
                       'encoders': encoded_data.encoders, 'is_multi_ts': is_multi_ts, 'stop_after': 1e2}
        if confidence_normalizer:
            normalizer = Normalizer(fit_params=norm_params)
            normalizer.fit(train_data)
            normalizer.prediction_cache = normalizer(encoded_data)
        else:
            normalizer = None

        # instance the ICP
        nc = nc_class(model, nc_function, normalizer=normalizer)
        icp = icp_class(nc)

        runtime_analyzer['icp']['__default'] = icp

        # setup prediction cache to avoid additional .predict() calls
        if is_classification:
            if predictor.models[predictor.best_index].supports_proba:
                icp.nc_function.model.prediction_cache = normal_predictions[all_cat_cols].values
            else:
                predicted_classes = pd.get_dummies(normal_predictions['prediction']).values  # inflate to one-hot enc
                icp.nc_function.model.prediction_cache = predicted_classes

        elif is_multi_ts:
            # we fit ICPs for time series confidence bounds only at t+1 forecast
            icp.nc_function.model.prediction_cache = np.array([p[0] for p in normal_predictions['prediction']])

        else:
            icp.nc_function.model.prediction_cache = np.array(normal_predictions['prediction'])

        if not is_classification:
            runtime_analyzer['df_std_dev'] = {'__default': stats_info.df_std_dev}

        # fit additional ICPs in time series tasks with grouped columns
        if ts_cfg.is_timeseries and ts_cfg.group_by:

            # create an ICP for each possible group
            group_info = data[ts_cfg.group_by].to_dict('list')
            all_group_combinations = list(product(*[set(x) for x in group_info.values()]))
            runtime_analyzer['icp']['__mdb_groups'] = all_group_combinations
            runtime_analyzer['icp']['__mdb_group_keys'] = [x for x in group_info.keys()]

            for combination in all_group_combinations:
                runtime_analyzer['icp'][frozenset(combination)] = deepcopy(icp)

        # calibrate ICP
        icp_df = deepcopy(data)
        icp_df, y = clean_df(icp_df, target, is_classification, runtime_analyzer.get('label_encoders', None))
        runtime_analyzer['icp']['__default'].index = icp_df.columns
        runtime_analyzer['icp']['__default'].calibrate(icp_df.values, y)

        # get confidence estimation for validation dataset
        conf, ranges = set_conf_range(
            icp_df, icp, dtype_dict[target],
            runtime_analyzer, positive_domain=positive_domain, significance=fixed_significance)
        if not is_classification:
            result_df = pd.DataFrame(index=data.index, columns=['confidence', 'lower', 'upper'], dtype=float)
            result_df.loc[icp_df.index, 'lower'] = ranges[:, 0]
            result_df.loc[icp_df.index, 'upper'] = ranges[:, 1]
        else:
            result_df = pd.DataFrame(index=data.index, columns=['confidence'], dtype=float)

        result_df.loc[icp_df.index, 'confidence'] = conf

        # calibrate additional grouped ICPs
        if ts_cfg.is_timeseries and ts_cfg.group_by:
            icps = runtime_analyzer['icp']
            group_keys = icps['__mdb_group_keys']

            # add all predictions to DF
            icps_df = deepcopy(data)
            if is_multi_ts:
                icps_df[f'__predicted_{target}'] = [p[0] for p in normal_predictions['prediction']]
            else:
                icps_df[f'__predicted_{target}'] = normal_predictions['prediction']

            for group in icps['__mdb_groups']:
                icp_df = icps_df
                if icps[frozenset(group)].nc_function.normalizer is not None:
                    icp_df[f'__norm_{target}'] = icps[frozenset(group)].nc_function.normalizer.prediction_cache

                # filter irrelevant rows for each group combination
                for key, val in zip(group_keys, group):
                    icp_df = icp_df[icp_df[key] == val]

                # save relevant predictions in the caches, then calibrate the ICP
                pred_cache = icp_df.pop(f'__predicted_{target}').values
                icps[frozenset(group)].nc_function.model.prediction_cache = pred_cache
                icp_df, y = clean_df(icp_df, target, is_classification, runtime_analyzer.get('label_encoders', None))
                if icps[frozenset(group)].nc_function.normalizer is not None:
                    icps[frozenset(group)].nc_function.normalizer.prediction_cache = icp_df.pop(
                        f'__norm_{target}').values

                icps[frozenset(group)].index = icp_df.columns      # important at inference time
                icps[frozenset(group)].calibrate(icp_df.values, y)

                # save training std() for bounds width selection
                if not is_classification:
                    icp_train_df = data
                    for key, val in zip(group_keys, group):
                        icp_train_df = icp_train_df[icp_train_df[key] == val]
                    y_train = icp_train_df[target].values
                    runtime_analyzer['df_std_dev'][frozenset(group)] = y_train.std()

                # get bounds for relevant rows in validation dataset
                conf, group_ranges = set_conf_range(
                    icp_df, icps[frozenset(group)],
                    dtype_dict[target],
                    runtime_analyzer, group=frozenset(group),
                    positive_domain=positive_domain, significance=fixed_significance)
                # save group bounds
                if not is_classification:
                    result_df.loc[icp_df.index, 'lower'] = group_ranges[:, 0]
                    result_df.loc[icp_df.index, 'upper'] = group_ranges[:, 1]

                result_df.loc[icp_df.index, 'confidence'] = conf

        # consolidate all groups here
        if not is_classification:
            ranges = result_df.values
            predictions['confidence_range'] = ranges

        runtime_analyzer['icp']['__mdb_active'] = True

    # get accuracy metric for validation data
    score_dict = evaluate_accuracy(
        data,
        normal_predictions['prediction'],
        target,
        accuracy_functions
    )
    normal_accuracy = np.mean(list(score_dict.values()))

    # compute global feature importance
    if not disable_column_importance:
        empty_input_accuracy = {}
        ignorable_input_cols = [x for x in input_cols if (not ts_cfg.is_timeseries or
                                                          (x not in ts_cfg.order_by and
                                                           x not in ts_cfg.historical_columns))]
        for col in ignorable_input_cols:
            partial_data = deepcopy(encoded_data)
            partial_data.clear_cache()
            for ds in partial_data.encoded_ds_arr:
                ds.data_frame[col] = [None] * len(ds.data_frame[col])

            if not is_classification:
                empty_input_preds = predictor(partial_data)
            else:
                empty_input_preds = predictor(partial_data, predict_proba=True)

            empty_input_accuracy[col] = np.mean(list(evaluate_accuracy(
                data,
                empty_input_preds['prediction'],
                target,
                accuracy_functions
            ).values()))

        column_importances = {}
        acc_increases = []
        for col in ignorable_input_cols:
            accuracy_increase = (normal_accuracy - empty_input_accuracy[col])
            acc_increases.append(accuracy_increase)

        # low 0.2 temperature to accentuate differences
        acc_increases = t_softmax(torch.Tensor([acc_increases]), t=0.2).tolist()[0]
        for col, inc in zip(ignorable_input_cols, acc_increases):
            column_importances[col] = 10 * inc  # scores go from 0 to 10 in GUI
    else:
        column_importances = None

    acc_stats = AccStats(dtype_dict=dtype_dict, target=target, buckets=stats_info.buckets)
    acc_stats.fit(data, normal_predictions, conf=result_df)
    bucket_accuracy, accuracy_histogram, cm, accuracy_samples = acc_stats.get_accuracy_stats(
        is_classification=is_classification, is_numerical=is_numerical)
    runtime_analyzer['bucket_accuracy'] = bucket_accuracy

    model_analysis = ModelAnalysis(
        accuracies=score_dict,
        accuracy_histogram=accuracy_histogram,
        accuracy_samples=accuracy_samples,
        train_sample_size=len(encoded_train_data),
        test_sample_size=len(encoded_data),
        confusion_matrix=cm,
        column_importances=column_importances,
        histograms=stats_info.histograms,
        dtypes=dtype_dict
    )

    return model_analysis, runtime_analyzer



from lightwood.model.unit import Unit
from lightwood.model.base import BaseModel
from lightwood.model.neural import Neural
from lightwood.model.lightgbm import LightGBM
from lightwood.model.lightgbm_array import LightGBMArray
from lightwood.model.sktime import SkTime
from lightwood.model.regression import Regression


__all__ = ['BaseModel', 'Neural', 'LightGBM', 'LightGBMArray', 'Unit', 'Regression', 'SkTime']



import math
import importlib
from typing import List, Union, Dict

import numpy as np
import pandas as pd
from sklearn.metrics import r2_score, f1_score, mean_absolute_error


def evaluate_accuracy(data: pd.DataFrame,
                      predictions: pd.Series,
                      target: str,
                      accuracy_functions: List[str]) -> Dict[str, float]:
    score_dict = {}

    for accuracy_function_str in accuracy_functions:
        if accuracy_function_str == 'evaluate_array_accuracy':
            nr_predictions = len(predictions.iloc[0])
            cols = [target] + [f'{target}_timestep_{i}' for i in range(1, nr_predictions)]
            true_values = data[cols].values.tolist()
            accuracy_function = evaluate_array_accuracy
        else:
            true_values = data[target].tolist()
            accuracy_function = getattr(importlib.import_module('sklearn.metrics'), accuracy_function_str)

        score_dict[accuracy_function_str] = accuracy_function(list(true_values), list(predictions))

    return score_dict


def evaluate_regression_accuracy(
        true_values,
        predictions,
        **kwargs
):
    if 'lower' and 'upper' in predictions:
        Y = np.array(true_values).astype(float)
        within = ((Y >= predictions['lower']) & (Y <= predictions['upper']))
        return sum(within) / len(within)
    else:
        r2 = r2_score(true_values, predictions['prediction'])
        return max(r2, 0)


def evaluate_multilabel_accuracy(true_values, predictions, **kwargs):
    pred_values = predictions['prediction']
    return f1_score(true_values, pred_values, average='weighted')


def evaluate_array_accuracy(
        true_values: List[List[Union[int, float]]],
        predictions: List[List[Union[int, float]]],
        **kwargs
) -> float:

    aggregate = 0

    for i in range(len(predictions)):
        try:
            valid_horizon = [math.isnan(x) for x in true_values[i]].index(True)
        except ValueError:
            valid_horizon = len(true_values[i])

        aggregate += mean_absolute_error(predictions[i][:valid_horizon],
                                         true_values[i][:valid_horizon])

    return aggregate / len(predictions)



class dtype():
    binary = 'binary'
    integer = 'integer'
    float = 'float'
    date = 'date'
    datetime = 'datetime'
    categorical = 'categorical'
    tags = 'tags'
    image = 'image'
    audio = 'audio'
    video = 'video'
    array = 'array'
    short_text = 'short_text'
    rich_text = 'rich_text'
    empty = 'empty'
    invalid = 'invalid'
    quantity = 'quantity'



'''
To: future intern reading this bit of code thinking "this is dumb, we could just check for isnan and isinf

No, no you can't, this will break subtly in a bunch of edge cases.

See the example (+ comments) in this article: https://cerebralab.com/Exceptions_as_control_flow

Sorry, python is dumb, but the spirit of your investigation wasn't, try again
'''


def can_be_nan_numeric(value):
    try:
        value = str(value)
        value = float(value)
    except Exception:
        return False

    try:
        if isinstance(value, float):
            a = int(value) # noqa
        isnan = False
    except Exception:
        isnan = True
    return isnan


def filter_nan(series):
    return [x for x in series if not can_be_nan_numeric(x)]



from copy import deepcopy
from lightwood.api.types import TimeseriesSettings
import re
from typing import Dict, List
from lightwood.api.dtype import dtype
from lightwood.helpers.log import log
from dateutil.parser import parse as parse_dt
import datetime
from lightwood.helpers.text import clean_float
import pandas as pd


def _to_datetime(element):
    try:
        date = parse_dt(element)
    except Exception:
        try:
            date = datetime.datetime.utcfromtimestamp(element)
        except Exception:
            return None

    return date


def _standardize_date(element):
    date = _to_datetime(element)
    if date is None:
        return None
    return date.timestamp()


def _standardize_datetime(element):
    date = _to_datetime(element)
    if date is None:
        return None
    return date.timestamp()


def _tags_to_tuples(tags_str):
    try:
        return tuple([x.strip() for x in tags_str.split(',')])
    except Exception:
        return tuple()


def _standardize_array(element):
    try:
        element = str(element)
        element = element.rstrip(']').lstrip('[')
        element = element.rstrip(' ').lstrip(' ')
        return element.replace(', ', ' ').replace(',', ' ')
    except Exception:
        return element


def _clean_float_or_none(element):
    try:
        return clean_float(element)
    except Exception:
        return None


def _clean_value(element: object, data_dtype: str):
    if data_dtype in (dtype.date):
        element = _standardize_date(element)

    if data_dtype in (dtype.datetime):
        element = _standardize_datetime(element)

    if data_dtype in (dtype.float):
        element = _clean_float_or_none(element)
    if data_dtype in (dtype.integer):
        element = int(_clean_float_or_none(element))

    if data_dtype in (dtype.array):
        element = _standardize_array(element)

    if data_dtype in (dtype.tags):
        element = _tags_to_tuples(element)

    if data_dtype in (dtype.quantity):
        element = float(re.sub("[^0-9.,]", '', element).replace(',', '.'))

    if data_dtype in (dtype.short_text, dtype.rich_text, dtype.categorical, dtype.binary):
        element = str(element)

    return element


def clean_empty_targets(df: pd.DataFrame, target: str) -> pd.DataFrame:
    len_before = len(df)
    df = df.dropna(subset=[target])
    len_after = len(df)
    nr_removed = len_before - len_after
    if nr_removed != 0:
        log.warning(
            f'Removed {nr_removed} rows due to the target value missing. Training with rows without a target value makes no sense, please avoid this!') # noqa

    return df


def cleaner(
        data: pd.DataFrame, dtype_dict: Dict[str, str],
        pct_invalid: float, ignore_features: List[str],
        identifiers: Dict[str, str],
        target: str, mode: str, timeseries_settings: TimeseriesSettings) -> pd.DataFrame:
    # Drop columns we don't want to use
    data = deepcopy(data)
    to_drop = [*ignore_features, *list(identifiers.keys())]
    exceptions = ['__mdb_make_predictions']
    for col in to_drop:
        try:
            data = data.drop(columns=[col])
        except Exception:
            pass

    if mode == 'train':
        data = clean_empty_targets(data, target)
    if mode == 'predict':
        if target in data.columns and not timeseries_settings.use_previous_target:
            data = data.drop(columns=[target])

    # Drop extra columns
    for name in list(data.columns):
        if name not in dtype_dict and name not in exceptions:
            data = data.drop(columns=[name])

    # Standardize content
    for name, data_dtype in dtype_dict.items():
        if mode == 'predict':
            if name == target:
                continue
        if name in to_drop:
            continue
        if name not in data.columns:
            if '__mdb_ts_previous' not in name:
                data[name] = [None] * len(data)
            continue

        new_data = []
        for element in data[name]:
            try:
                new_data.append(_clean_value(element, data_dtype))
            except Exception as e:
                new_data.append(None)
                log.warning(
                    f'Unable to parse elemnt: {element} or type {data_dtype} from column {name}. Excetpion: {e}')

        pct_invalid = 100 * (len(new_data) - len([x for x in new_data if x is not None])) / len(new_data)

        if pct_invalid > pct_invalid:
            err = f'Too many ({pct_invalid}%) invalid values in column {name} of type {data_dtype}'
            log.error(err)
            raise Exception(err)

        data[name] = new_data

    return data



from functools import partial

# Factories are syntactic sugar but can introduce bugs, we should *only* use them for the public APIs


class Factory:
    self.generate_class = None

    def __init__(self, **kwargs):
        pass

    def generate(self) -> self.generate_class:
        return self.generate_class(**kwargs)


def gen_factory_func(func: Callable, **kwargs) -> Callable:
    return partial(func, **kwargs)



import pandas as pd
import numpy as np
from typing import List
from itertools import product

from lightwood.api.types import TimeseriesSettings


def splitter(data: pd.DataFrame, k: int, tss: TimeseriesSettings) -> List[pd.DataFrame]:
    if not tss.is_timeseries:
        # shuffle
        data = data.sample(frac=1).reset_index(drop=True)

        # split
        folds = np.array_split(data, k)

    else:
        if not tss.group_by:
            folds = np.array_split(data, k)
        else:
            gcols = tss.group_by
            all_group_combinations = list(product(*[data[gcol].unique() for gcol in tss.group_by]))
            folds = [pd.DataFrame() for _ in range(k)]
            for group in all_group_combinations:
                subframe = data
                for idx, gcol in enumerate(gcols):
                    subframe = subframe[subframe[gcol] == group[idx]]

                subfolds = np.array_split(subframe, k)

                for i in range(k):
                    folds[i] = pd.concat([folds[i], subfolds[i]])

    return folds



from copy import deepcopy

import torch
import numpy as np
import pandas as pd

# restore_icp_state, clear_icp_state
from lightwood.analysis.nc.util import get_numerical_conf_range, get_categorical_conf, get_anomalies
from lightwood.helpers.ts import get_inferred_timestamps, add_tn_conf_bounds
from lightwood.api.dtype import dtype
from lightwood.api.types import TimeseriesSettings


def explain(data: pd.DataFrame,
            encoded_data: torch.Tensor,
            predictions: pd.DataFrame,
            timeseries_settings: TimeseriesSettings,
            analysis: dict,
            target_name: str,
            target_dtype: str,
            positive_domain: bool,
            fixed_confidence: float,
            anomaly_detection: bool,

            # forces specific confidence level in ICP
            anomaly_error_rate: float,

            # ignores anomaly detection for N steps after an
            # initial anomaly triggers the cooldown period;
            # implicitly assumes series are regularly spaced
            anomaly_cooldown: int,

            ts_analysis: dict = None
            ):

    # @TODO: check not quick_predict
    data = data.reset_index(drop=True)

    insights = pd.DataFrame()
    if target_name in data.columns:
        insights['truth'] = data[target_name]
    else:
        insights['truth'] = [None] * len(predictions['prediction'])
    insights['prediction'] = predictions['prediction']

    if timeseries_settings.is_timeseries:

        if timeseries_settings.group_by:
            for col in timeseries_settings.group_by:
                insights[f'group_{col}'] = data[col]

        for col in timeseries_settings.order_by:
            insights[f'order_{col}'] = data[col]

        for col in timeseries_settings.order_by:
            insights[f'order_{col}'] = get_inferred_timestamps(
                insights, col, ts_analysis['deltas'], timeseries_settings)

    # confidence estimation using calibrated inductive conformal predictors (ICPs)
    if analysis['icp']['__mdb_active']:
        icp_X = deepcopy(data)

        # replace observed data w/predictions
        preds = predictions['prediction']
        if timeseries_settings.is_timeseries and timeseries_settings.nr_predictions > 1:
            preds = [p[0] for p in preds]

            for col in [f'timestep_{i}' for i in range(1, timeseries_settings.nr_predictions)]:
                if col in icp_X.columns:
                    icp_X.pop(col)  # erase ignorable columns

        icp_X[target_name] = preds

        is_categorical = target_dtype in (dtype.binary, dtype.categorical, dtype.array)
        is_numerical = target_dtype in [dtype.integer, dtype.float] or target_dtype == dtype.array
        is_anomaly_task = is_numerical and timeseries_settings.is_timeseries and anomaly_detection

        if (is_numerical or is_categorical) and analysis['icp'].get('__mdb_active', False):

            # reorder DF index
            index = analysis['icp']['__default'].index.values
            index = np.append(index, target_name) if target_name not in index else index
            icp_X = icp_X.reindex(columns=index)  # important, else bounds can be invalid

            # only one normalizer, even if it's a grouped time series task
            normalizer = analysis['icp']['__default'].nc_function.normalizer
            if normalizer:
                normalizer.prediction_cache = normalizer(encoded_data)
                icp_X['__mdb_selfaware_scores'] = normalizer.prediction_cache

            # get ICP predictions
            result_cols = ['lower', 'upper', 'significance'] if is_numerical else ['significance']
            result = pd.DataFrame(index=icp_X.index, columns=result_cols)

            # base ICP
            X = deepcopy(icp_X)

            # get all possible ranges
            if timeseries_settings.is_timeseries and timeseries_settings.nr_predictions > 1 and is_numerical:

                # bounds in time series are only given for the first forecast
                analysis['icp']['__default'].nc_function.model.prediction_cache = \
                    [p[0] for p in predictions['prediction']]
                all_confs = analysis['icp']['__default'].predict(X.values)

            elif is_numerical:
                analysis['icp']['__default'].nc_function.model.prediction_cache = predictions['prediction']
                all_confs = analysis['icp']['__default'].predict(X.values)

            # categorical
            else:
                predicted_proba = True if any(['__mdb_proba' in col for col in predictions.columns]) else False
                if predicted_proba:
                    all_cat_cols = [col for col in predictions.columns if '__mdb_proba' in col]
                    class_dists = predictions[all_cat_cols].values
                    for icol, cat_col in enumerate(all_cat_cols):
                        insights.loc[X.index, cat_col] = class_dists[:, icol]
                else:
                    class_dists = pd.get_dummies(predictions['prediction']).values

                analysis['icp']['__default'].nc_function.model.prediction_cache = class_dists

                conf_candidates = list(range(20)) + list(range(20, 100, 10))
                all_ranges = np.array(
                    [analysis['icp']['__default'].predict(X.values, significance=s / 100)
                     for s in conf_candidates])
                all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)

            # convert (B, 2, 99) into (B, 2) given width or error rate constraints
            if is_numerical:
                significances = fixed_confidence
                if significances is not None:
                    confs = all_confs[:, :, int(100 * (1 - significances)) - 1]
                else:
                    error_rate = anomaly_error_rate if is_anomaly_task else None
                    significances, confs = get_numerical_conf_range(all_confs,
                                                                    df_std_dev=analysis['df_std_dev'],
                                                                    positive_domain=positive_domain,
                                                                    error_rate=error_rate)
                result.loc[X.index, 'lower'] = confs[:, 0]
                result.loc[X.index, 'upper'] = confs[:, 1]
            else:
                conf_candidates = list(range(20)) + list(range(20, 100, 10))
                significances = get_categorical_conf(all_confs, conf_candidates)

            result.loc[X.index, 'significance'] = significances

            # grouped time series, we replace bounds in rows that have a trained ICP
            if analysis['icp'].get('__mdb_groups', False):
                icps = analysis['icp']
                group_keys = icps['__mdb_group_keys']

                for group in icps['__mdb_groups']:
                    icp = icps[frozenset(group)]

                    # check ICP has calibration scores
                    if icp.cal_scores[0].shape[0] > 0:

                        # filter rows by group
                        X = deepcopy(icp_X)
                        for key, val in zip(group_keys, group):
                            X = X[X[key] == val]

                        if X.size > 0:
                            # set ICP caches
                            icp.nc_function.model.prediction_cache = X.pop(target_name).values
                            if icp.nc_function.normalizer:
                                icp.nc_function.normalizer.prediction_cache = X.pop('__mdb_selfaware_scores').values

                            # predict and get confidence level given width or error rate constraints
                            if is_numerical:
                                all_confs = icp.predict(X.values)
                                error_rate = anomaly_error_rate if is_anomaly_task else None
                                significances, confs = get_numerical_conf_range(all_confs,
                                                                                df_std_dev=analysis['df_std_dev'],
                                                                                positive_domain=positive_domain,
                                                                                group=frozenset(group),
                                                                                error_rate=error_rate)

                                # only replace where grouped ICP is more informative (i.e. tighter)
                                default_icp_widths = result.loc[X.index, 'upper'] - result.loc[X.index, 'lower']
                                grouped_widths = np.subtract(confs[:, 1], confs[:, 0])
                                insert_index = (default_icp_widths > grouped_widths)[lambda x: x.isin([True])].index
                                conf_index = (default_icp_widths.reset_index(drop=True) >
                                              grouped_widths)[lambda x: x.isin([True])].index

                                result.loc[insert_index, 'lower'] = confs[conf_index, 0]
                                result.loc[insert_index, 'upper'] = confs[conf_index, 1]
                                result.loc[insert_index, 'significance'] = significances[conf_index]

                            else:
                                conf_candidates = list(range(20)) + list(range(20, 100, 10))
                                all_ranges = np.array(
                                    [icp.predict(X.values, significance=s / 100)
                                     for s in conf_candidates])
                                all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
                                significances = get_categorical_conf(all_confs, conf_candidates)
                                result.loc[X.index, 'significance'] = significances

            insights['confidence'] = result['significance'].astype(float).tolist()

            if is_numerical:
                insights['lower'] = result['lower'].astype(float)
                insights['upper'] = result['upper'].astype(float)

            # anomaly detection
            if is_anomaly_task:
                anomalies = get_anomalies(insights,
                                          data[target_name],
                                          cooldown=anomaly_cooldown)
                insights['anomaly'] = anomalies

            if timeseries_settings.is_timeseries and timeseries_settings.nr_predictions > 1 and is_numerical:
                insights = add_tn_conf_bounds(insights, timeseries_settings)

    return insights



from lightwood.api.types import ModelAnalysis
import dill
import pandas as pd


# Interface that must be respected by predictor objects generated from JSON ML and/or compatible with Mindsdb
class PredictorInterface():
    model_analysis: ModelAnalysis = None

    def __init__(self):
        pass

    def learn(self, data: pd.DataFrame) -> None:
        pass

    def adjust(self, data: pd.DataFrame) -> None:
        pass

    def predict(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    def predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    def save(self, file_path: str):
        with open(file_path, 'wb') as fp:
            dill.dump(self, fp)



from lightwood.api.dtype import dtype
from lightwood.api.types import (JsonAI, Output, Feature, TypeInformation, StatisticalAnalysis, ProblemDefinition,
                                 TimeseriesSettings, ModelAnalysis, DataAnalysis)
from lightwood.api.predictor import PredictorInterface
from lightwood.api.encode import encode
from lightwood.api.high_level import (analyze_dataset, code_from_problem, predictor_from_problem, predictor_from_code,
                                      code_from_json_ai, json_ai_from_problem, predictor_from_state)

__all__ = [
    'analyze_dataset', 'code_from_problem', 'predictor_from_problem', 'predictor_from_code', 'code_from_json_ai',
    'json_ai_from_problem', 'JsonAI', 'Output', 'Feature', 'TypeInformation', 'StatisticalAnalysis',
    'ProblemDefinition', 'TimeseriesSettings', 'ModelAnalysis', 'DataAnalysis', 'PredictorInterface', 'encode',
    'dtype', 'predictor_from_state']



from typing import Dict
import pandas as pd
import numpy as np

from lightwood.api import StatisticalAnalysis, ProblemDefinition
from lightwood.helpers.numeric import filter_nan
from lightwood.helpers.seed import seed
from lightwood.data.cleaner import cleaner
from lightwood.helpers.log import log
from lightwood.api.dtype import dtype
from scipy.stats import entropy
from lightwood.data.cleaner import _clean_float_or_none


def get_numeric_histogram(data, data_dtype):
    data = [_clean_float_or_none(x) for x in data]

    Y, X = np.histogram(data, bins=min(50, len(set(data))),
                        range=(min(data), max(data)), density=False)
    if data_dtype == dtype.integer:
        Y, X = np.histogram(data, bins=[int(round(x)) for x in X], density=False)

    X = X[:-1].tolist()
    Y = Y.tolist()

    return {
        'x': X,
        'y': Y
    }


def compute_entropy_biased_buckets(histogram):
    S, biased_buckets = None, None
    if histogram is not None:
        hist_x = histogram['x']
        hist_y = histogram['y']
        nr_values = sum(hist_y)
        S = entropy([x / nr_values for x in hist_y], base=max(2, len(hist_y)))
        if S < 0.25:
            pick_nr = -max(1, int(len(hist_y) / 10))
            biased_buckets = [hist_x[i] for i in np.array(hist_y).argsort()[pick_nr:]]
    return S, biased_buckets


def statistical_analysis(data: pd.DataFrame,
                         dtypes: Dict[str, str],
                         identifiers: Dict[str, object],
                         problem_definition: ProblemDefinition) -> StatisticalAnalysis:
    seed()
    log.info('Starting statistical analysis')
    df = cleaner(data, dtypes, problem_definition.pct_invalid, problem_definition.ignore_features,
                 identifiers, problem_definition.target, 'train', problem_definition.timeseries_settings)

    missing = {col: len([x for x in df[col] if x is None]) / len(df[col]) for col in df.columns}
    distinct = {col: len(set(df[col])) / len(df[col]) for col in df.columns}

    nr_rows = len(df)
    target = problem_definition.target
    # get train std, used in analysis
    if dtypes[target] in [dtype.float, dtype.integer, dtype.array]:
        df_std = df[target].astype(float).std()
    elif dtypes[target] in [dtype.array]:
        try:
            all_vals = []
            for x in df[target]:
                all_vals += x
            df_std = pd.Series(all_vals).astype(float).std()
        except Exception as e:
            log.warning(e)
            df_std = 1.0
    else:
        df_std = 1.0

    histograms = {}
    buckets = {}
    # Get histograms for each column
    for col in df.columns:
        histograms[col] = None
        buckets[col] = None
        if dtypes[col] in (dtype.categorical, dtype.binary):
            hist = dict(df[col].value_counts().apply(lambda x: x / len(df[col])))
            histograms[col] = {
                'x': list(hist.keys()),
                'y': list(hist.values())
            }
            buckets[col] = histograms[col]['x']
        if dtypes[col] in (dtype.integer, dtype.float, dtype.array):
            histograms[col] = get_numeric_histogram(filter_nan(df[col]), dtypes[col])
            buckets[col] = histograms[col]['x']

    # get observed classes, used in analysis
    target_class_distribution = None
    if dtypes[target] in (dtype.categorical, dtype.binary):
        target_class_distribution = dict(df[target].value_counts().apply(lambda x: x / len(df[target])))
        train_observed_classes = list(target_class_distribution.keys())
    elif dtypes[target] == dtype.tags:
        train_observed_classes = None  # @TODO: pending call to tags logic -> get all possible tags
    else:
        train_observed_classes = None

    bias = {}
    for col in df.columns:
        S, biased_buckets = compute_entropy_biased_buckets(histograms[col])
        bias[col] = {
            'entropy': S,
            'description': """Under the assumption of uniformly distributed data (i.e., same probability for Head or Tails on a coin flip) mindsdb tries to detect potential divergences from such case, and it calls this "potential bias". Thus by our data having any potential bias mindsdb means any divergence from all categories having the same probability of being selected.""", # noqa
            'biased_buckets': biased_buckets
        }

    avg_words_per_sentence = {}
    for col in df.columns:
        if dtypes[col] in (dtype.rich_text, dtype.short_text):
            words_per_sentence = []
            for item in df[col]:
                words_per_sentence.append(len(item.split(' ')))
            avg_words_per_sentence[col] = int(np.mean(words_per_sentence))
        else:
            avg_words_per_sentence[col] = None

    log.info('Finished statistical analysis')
    return StatisticalAnalysis(
        nr_rows=nr_rows,
        df_std_dev=df_std,
        train_observed_classes=train_observed_classes,
        target_class_distribution=target_class_distribution,
        histograms=histograms,
        buckets=buckets,
        missing=missing,
        distinct=distinct,
        bias=bias,
        avg_words_per_sentence=avg_words_per_sentence
    )



from typing import Dict
import pandas as pd

from lightwood.api.types import TimeseriesSettings
from lightwood.encoder.time_series.helpers.common import get_group_matches, generate_target_group_normalizers


def timeseries_analyzer(data: pd.DataFrame, dtype_dict: Dict[str, str],
                        timeseries_settings: TimeseriesSettings, target: str) -> (Dict, Dict):
    info = {
        'original_type': dtype_dict[target],
        'data': data[target].values
    }
    if timeseries_settings.group_by is not None:
        info['group_info'] = {gcol: data[gcol].tolist() for gcol in timeseries_settings.group_by}  # group col values
    else:
        info['group_info'] = {}

    # @TODO: maybe normalizers should fit using only the training folds??
    new_data = generate_target_group_normalizers(info)

    deltas = get_delta(data[timeseries_settings.order_by],
                       info,
                       new_data['group_combinations'],
                       timeseries_settings.order_by)

    return {'target_normalizers': new_data['target_normalizers'],
            'deltas': deltas,
            'tss': timeseries_settings,
            'group_combinations': new_data['group_combinations']}


def get_delta(df: pd.DataFrame, ts_info: dict, group_combinations: list, order_cols: list):
    """
    Infer the sampling interval of each time series
    """
    deltas = {"__default": {}}

    for col in order_cols:
        series = pd.Series([x[-1] for x in df[col]])
        rolling_diff = series.rolling(window=2).apply(lambda x: x.iloc[1] - x.iloc[0])
        delta = rolling_diff.value_counts(ascending=False).keys()[0]
        deltas["__default"][col] = delta

    if ts_info.get('group_info', False):
        for group in group_combinations:
            if group != "__default":
                deltas[group] = {}
                for col in order_cols:
                    ts_info['data'] = pd.Series([x[-1] for x in df[col]])
                    _, subset = get_group_matches(ts_info, group)
                    if subset.size > 1:
                        rolling_diff = pd.Series(
                            subset.squeeze()).rolling(
                            window=2).apply(
                            lambda x: x.iloc[1] - x.iloc[0])
                        delta = rolling_diff.value_counts(ascending=False).keys()[0]
                        deltas[group][col] = delta

    return deltas



import os
import torch
from torch.nn.functional import softmax
from lightwood.analysis.nc.base import RegressorAdapter, ClassifierAdapter


def t_softmax(x, t=1.0, axis=1):
    """ Softmax with temperature scaling """
    # @TODO: move this, not a wrapper
    return softmax(torch.Tensor(x) / t, dim=axis).numpy()


def clear_icp_state(icp):
    """ We clear last_x and last_y to minimize file size. Model has to be cleared because it cannot be pickled. """
    icp.model.model = None
    icp.model.last_x = None
    icp.model.last_y = None
    if icp.normalizer is not None:
        icp.normalizer.model = None


def restore_icp_state(col, hmd, session):
    icps = hmd['icp'][col]
    try:
        predictor = session.transaction.model_backend.predictor
    except AttributeError:
        model_path = os.path.join(CONFIG.MINDSDB_STORAGE_PATH, hmd['name'], 'lightwood_data')
        predictor = Predictor(load_from_path=model_path)

    for group, icp in icps.items():
        if group not in ['__mdb_groups', '__mdb_group_keys']:
            icp.nc_function.model.model = predictor

    # restore model in normalizer
    for group, icp in icps.items():
        if group not in ['__mdb_groups', '__mdb_group_keys']:
            if icp.nc_function.normalizer is not None:
                icp.nc_function.normalizer.model = icp.nc_function.model.model


class ConformalRegressorAdapter(RegressorAdapter):
    def __init__(self, model, fit_params=None):
        super(ConformalRegressorAdapter, self).__init__(model, fit_params)
        self.prediction_cache = None

    def fit(self, x=None, y=None):
        """ At this point, the predictor has already been trained, but this
        has to be called to setup some things in the nonconformist backend """
        pass

    def predict(self, x=None):
        """ Same as in .fit()
        :return: np.array (n_test, n_classes) with class probability estimates """
        return self.prediction_cache


class ConformalClassifierAdapter(ClassifierAdapter):
    def __init__(self, model, fit_params=None):
        super(ConformalClassifierAdapter, self).__init__(model, fit_params)
        self.prediction_cache = None

    def fit(self, x=None, y=None):
        """ At this point, the predictor has already been trained, but this
        has to be called to setup some things in the nonconformist backend """
        pass

    def predict(self, x=None):
        """ Same as in .fit()
        :return: np.array (n_test, n_classes) with class probability estimates """
        return t_softmax(self.prediction_cache, t=0.5)



from typing import List
from lightwood.model.base import BaseModel
import pandas as pd
from lightwood.data.encoded_ds import EncodedDs


class BaseEnsemble:
    data: List[EncodedDs]
    models: List[BaseModel]
    best_index: int
    supports_proba: bool

    def __init__(self, target, models: List[BaseModel], data: List[EncodedDs]) -> None:
        self.data = data
        self.models = models
        self.best_index = 0
        self.supports_proba = False

    def __call__(self, ds: EncodedDs, predict_proba: bool = False) -> pd.DataFrame:
        raise NotImplementedError()



from typing import Dict, List, Optional, Union
from dataclasses import dataclass
from lightwood.helpers.log import log
from dataclasses_json import dataclass_json
from dataclasses_json.core import _asdict, Json
import json


@dataclass
class Feature:
    encoder: str
    data_dtype: str = None
    dependency: List[str] = None

    @staticmethod
    def from_dict(obj: Dict):
        encoder = obj['encoder']
        data_dtype = obj.get('data_dtype', None)
        dependency = obj.get('dependency', None)

        feature = Feature(
            encoder=encoder,
            data_dtype=data_dtype,
            dependency=dependency
        )

        return feature

    @staticmethod
    def from_json(data: str):
        return Feature.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        as_dict = _asdict(self, encode_json=encode_json)
        for k in list(as_dict.keys()):
            if as_dict[k] is None:
                del as_dict[k]
        return as_dict

    def to_json(self) -> Dict[str, Json]:
        return json.dumps(self.to_dict(), indent=4)


@dataclass_json
@dataclass
class Output:
    data_dtype: str
    encoder: str = None
    models: List[str] = None
    ensemble: str = None


@dataclass_json
@dataclass
class TypeInformation:
    dtypes: Dict[str, str]
    additional_info: Dict[str, object]
    identifiers: Dict[str, str]

    def __init__(self):
        self.dtypes = dict()
        self.additional_info = dict()
        self.identifiers = dict()


@dataclass_json
@dataclass
class StatisticalAnalysis:
    nr_rows: int
    df_std_dev: Optional[float]
    # Write proper to and from dict parsing for this than switch back to using the types bellow, dataclasses_json sucks!
    train_observed_classes: object  # Union[None, List[str]]
    target_class_distribution: object  # Dict[str, float]
    histograms: object  # Dict[str, Dict[str, List[object]]]
    buckets: object  # Dict[str, Dict[str, List[object]]]
    missing: object
    distinct: object
    bias: object
    avg_words_per_sentence: object


@dataclass_json
@dataclass
class DataAnalysis:
    statistical_analysis: StatisticalAnalysis
    type_information: TypeInformation


@dataclass
class TimeseriesSettings:
    is_timeseries: bool
    order_by: List[str] = None
    window: int = None
    group_by: List[str] = None
    use_previous_target: bool = False
    nr_predictions: int = None
    historical_columns: List[str] = None
    target_type: str = ''  # @TODO: is the current setter (outside of initialization) a sane option?

    @staticmethod
    def from_dict(obj: Dict):
        if len(obj) > 0:
            for mandatory_setting in ['order_by', 'window']:
                if mandatory_setting not in obj:
                    err = f'Missing mandatory timeseries setting: {mandatory_setting}'
                    log.error(err)
                    raise Exception(err)

            timeseries_settings = TimeseriesSettings(
                is_timeseries=True,
                order_by=obj['order_by'],
                window=obj['window'],
                use_previous_target=obj.get('use_previous_target', True),
                historical_columns=[],
                nr_predictions=obj.get('nr_predictions', 1)
            )
            for setting in obj:
                timeseries_settings.__setattr__(setting, obj[setting])

        else:
            timeseries_settings = TimeseriesSettings(is_timeseries=False)

        return timeseries_settings

    @staticmethod
    def from_json(data: str):
        return TimeseriesSettings.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        return _asdict(self, encode_json=encode_json)

    def to_json(self) -> Dict[str, Json]:
        return json.dumps(self.to_dict())


@dataclass
class ProblemDefinition:
    target: str
    nfolds: int
    pct_invalid: float
    unbias_target: bool
    seconds_per_model: Union[int, None]
    seconds_per_encoder: Union[int, None]
    time_aim: Union[int, None]
    target_weights: Union[List[float], None]
    positive_domain: bool
    fixed_confidence: Union[int, float, None]
    timeseries_settings: TimeseriesSettings
    anomaly_detection: bool
    anomaly_error_rate: Union[float, None]
    anomaly_cooldown: int
    ignore_features: List[str]
    fit_on_validation: bool
    strict_mode: bool

    @staticmethod
    def from_dict(obj: Dict):
        target = obj['target']
        nfolds = obj.get('nfolds', 30)
        pct_invalid = obj.get('pct_invalid', 1)
        unbias_target = obj.get('unbias_target', False)
        seconds_per_model = obj.get('seconds_per_model', None)
        seconds_per_encoder = obj.get('seconds_per_encoder', None)
        time_aim = obj.get('time_aim', None)
        target_weights = obj.get('target_weights', None)
        positive_domain = obj.get('positive_domain', False)
        fixed_confidence = obj.get('fixed_confidence', None)
        timeseries_settings = TimeseriesSettings.from_dict(obj.get('timeseries_settings', {}))
        anomaly_detection = obj.get('anomaly_detection', True)
        anomaly_error_rate = obj.get('anomaly_error_rate', None)
        anomaly_cooldown = obj.get('anomaly_detection', 1)
        ignore_features = obj.get('ignore_features', [])
        fit_on_validation = obj.get('fit_on_validation', True)
        strict_mode = obj.get('strict_mode', True)

        problem_definition = ProblemDefinition(
            target=target,
            nfolds=nfolds,
            pct_invalid=pct_invalid,
            unbias_target=unbias_target,
            seconds_per_model=seconds_per_model,
            seconds_per_encoder=seconds_per_encoder,
            time_aim=time_aim,
            target_weights=target_weights,
            positive_domain=positive_domain,
            fixed_confidence=fixed_confidence,
            timeseries_settings=timeseries_settings,
            anomaly_detection=anomaly_detection,
            anomaly_error_rate=anomaly_error_rate,
            anomaly_cooldown=anomaly_cooldown,
            ignore_features=ignore_features,
            fit_on_validation=fit_on_validation,
            strict_mode=strict_mode
        )

        return problem_definition

    @staticmethod
    def from_json(data: str):
        return ProblemDefinition.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        return _asdict(self, encode_json=encode_json)

    def to_json(self) -> Dict[str, Json]:
        return json.dumps(self.to_dict())


@dataclass
class JsonAI:
    features: Dict[str, Feature]
    outputs: Dict[str, Output]
    problem_definition: ProblemDefinition
    identifiers: Dict[str, str]
    cleaner: Optional[object] = None
    splitter: Optional[object] = None
    analyzer: Optional[object] = None
    explainer: Optional[object] = None
    imports: Optional[List[str]] = None
    timeseries_transformer: Optional[object] = None
    timeseries_analyzer: Optional[object] = None
    accuracy_functions: Optional[List[str]] = None
    phases: Optional[Dict[str, object]] = None

    @staticmethod
    def from_dict(obj: Dict):
        features = {k: Feature.from_dict(v) for k, v in obj['features'].items()}
        outputs = {k: Output.from_dict(v) for k, v in obj['outputs'].items()}
        problem_definition = ProblemDefinition.from_dict(obj['problem_definition'])
        identifiers = obj['identifiers']
        cleaner = obj.get('cleaner', None)
        splitter = obj.get('splitter', None)
        analyzer = obj.get('analyzer', None)
        explainer = obj.get('explainer', None)
        imports = obj.get('imports', None)
        timeseries_transformer = obj.get('timeseries_transformer', None)
        timeseries_analyzer = obj.get('timeseries_analyzer', None)
        accuracy_functions = obj.get('accuracy_functions', None)
        phases = obj.get('phases', None)

        json_ai = JsonAI(
            features=features,
            outputs=outputs,
            problem_definition=problem_definition,
            identifiers=identifiers,
            cleaner=cleaner,
            splitter=splitter,
            analyzer=analyzer,
            explainer=explainer,
            imports=imports,
            timeseries_transformer=timeseries_transformer,
            timeseries_analyzer=timeseries_analyzer,
            accuracy_functions=accuracy_functions,
            phases=phases
        )

        return json_ai

    @staticmethod
    def from_json(data: str):
        return JsonAI.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        as_dict = _asdict(self, encode_json=encode_json)
        for k in list(as_dict.keys()):
            if k == 'features':
                feature_dict = {}
                for name in self.features:
                    feature_dict[name] = self.features[name].to_dict()
                as_dict[k] = feature_dict
            if as_dict[k] is None:
                del as_dict[k]
        return as_dict

    def to_json(self) -> Dict[str, Json]:
        return json.dumps(self.to_dict(), indent=4)


@dataclass_json
@dataclass
class ModelAnalysis:
    accuracies: Dict[str, float]
    accuracy_histogram: Dict[str, list]
    accuracy_samples: Dict[str, list]
    train_sample_size: int
    test_sample_size: int
    column_importances: Dict[str, float]
    confusion_matrix: object
    histograms: object
    dtypes: object



import os
import dill
import pandas as pd
from lightwood.api.types import DataAnalysis, JsonAI, ProblemDefinition
import lightwood
from lightwood.api.predictor import PredictorInterface
from lightwood.api.json_ai import generate_json_ai
import tempfile
from lightwood.api.json_ai import code_from_json_ai as _code_from_json_ai
import importlib.util
import sys
import random
import string
import gc
import time


def _module_from_code(code, module_name):
    dirname = tempfile.gettempdir()
    filename = os.urandom(24).hex() + str(time.time()).replace('.', '') + '.py'
    path = os.path.join(dirname, filename)
    if 'LIGHTWOOD_DEV_SAVE_TO' in os.environ:
        path = os.environ['LIGHTWOOD_DEV_SAVE_TO']

    with open(path, 'wb') as fp:
        fp.write(code.encode('utf-8'))
        spec = importlib.util.spec_from_file_location(module_name, fp.name)
        temp_module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = temp_module
        spec.loader.exec_module(temp_module)

    return temp_module


def code_from_json_ai(json_ai: JsonAI) -> str:
    return _code_from_json_ai(json_ai)


def analyze_dataset(df: pd.DataFrame) -> DataAnalysis:
    # @TODO Add missing value % and warning flags
    problem_definition = ProblemDefinition.from_dict({'target': str(df.columns[0])})

    type_information = lightwood.data.infer_types(df, problem_definition.pct_invalid)
    statistical_analysis = lightwood.data.statistical_analysis(
        df, type_information.dtypes, type_information.identifiers, problem_definition)

    return DataAnalysis(
        type_information=type_information,
        statistical_analysis=statistical_analysis
    )


def json_ai_from_problem(df: pd.DataFrame, problem_definition: ProblemDefinition) -> JsonAI:
    if not isinstance(problem_definition, ProblemDefinition):
        problem_definition = ProblemDefinition.from_dict(problem_definition)

    type_information = lightwood.data.infer_types(df, problem_definition.pct_invalid)
    statistical_analysis = lightwood.data.statistical_analysis(
        df, type_information.dtypes, type_information.identifiers, problem_definition)
    json_ai = generate_json_ai(
        type_information=type_information, statistical_analysis=statistical_analysis,
        problem_definition=problem_definition)

    return json_ai


def code_from_problem(df: pd.DataFrame, problem_definition: ProblemDefinition) -> str:
    json_ai = json_ai_from_problem(df, problem_definition)
    predictor_code = code_from_json_ai(json_ai)
    return predictor_code


def predictor_from_code(code: str) -> PredictorInterface:
    module_name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=12))
    module_name += str(time.time()).replace('.', '')
    predictor = _module_from_code(code, module_name).Predictor()
    return predictor


def predictor_from_state(state_file: str, code: str = None) -> PredictorInterface:
    try:
        module_name = None
        with open(state_file, 'rb') as fp:
            predictor = dill.load(fp)
    except Exception as e:
        module_name = str(e).lstrip("No module named '").split("'")[0]
        if code is None:
            raise Exception(
                'Provide code when loading a predictor from outside the scope/script it was created in!')

    if module_name is not None:
        try:
            del sys.modules[module_name]
        except Exception:
            pass
        gc.collect()
        _module_from_code(code, module_name)
        with open(state_file, 'rb') as fp:
            predictor = dill.load(fp)

    return predictor


def predictor_from_problem(df: pd.DataFrame, problem_definition: ProblemDefinition) -> PredictorInterface:
    if not isinstance(problem_definition, ProblemDefinition):
        problem_definition = ProblemDefinition.from_dict(problem_definition)

    predictor_class_str = code_from_problem(df, problem_definition)
    return predictor_from_code(predictor_class_str)



from typing import List

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.model.base import BaseModel
from lightwood.ensemble.base import BaseEnsemble
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs
from lightwood.helpers.general import evaluate_accuracy


class BestOf(BaseEnsemble):
    best_index: int

    def __init__(self, target, models: List[BaseModel], data: List[EncodedDs], accuracy_functions) -> None:
        super().__init__(target, models, data)
        # @TODO: Need some shared accuracy functionality to determine model selection here
        self.maximize = False if 'evaluate_array_accuracy' in accuracy_functions else True
        best_score = -pow(2, 32) if self.maximize else pow(2, 32)
        ds = ConcatedEncodedDs(data)
        for idx, model in enumerate(models):
            score_dict = evaluate_accuracy(
                ds.data_frame,
                model(ds)['prediction'],
                target,
                accuracy_functions
            )
            avg_score = np.mean(list(score_dict.values()))
            log.info(f'Model {type(model).__name__} obtained a best-of evaluation score of {round(avg_score,4)}')
            if self.improves(avg_score, best_score, accuracy_functions):
                best_score = avg_score
                self.best_index = idx

        self.supports_proba = self.models[self.best_index].supports_proba
        log.info(f'Picked best model: {type(self.models[self.best_index]).__name__}')

    def __call__(self, ds: EncodedDs, predict_proba: bool = False) -> pd.DataFrame:
        return self.models[self.best_index](ds, predict_proba=predict_proba)

    def improves(self, new, old, functions):
        return new > old if self.maximize else new < old



import unittest
import numpy as np
import pandas as pd
from typing import List

from lightwood.api.types import ProblemDefinition

np.random.seed(0)


class TestTimeseries(unittest.TestCase):
    def check_ts_prediction_df(self, df: pd.DataFrame, nr_preds: int, orders: List[str]):
        for idx, row in df.iterrows():
            assert len(row['prediction']) == nr_preds

            for oby in orders:
                assert len(row[f'order_{oby}']) == nr_preds

            for t in range(nr_preds):
                assert row['lower'][t] <= row['prediction'][t] <= row['upper'][t]

            for oby in orders:
                assert len(row[f'order_{oby}']) == nr_preds

            if row['anomaly']:
                assert not (row['lower'][0] <= row['truth'] <= row['upper'][0])
            else:
                assert row['lower'][0] <= row['truth'] <= row['upper'][0]

    def test_0_grouped_regression_timeseries(self):
        """ Test grouped numerical predictions (forecast horizon > 1), covering most of the TS pipeline """
        from lightwood.api.high_level import predictor_from_problem

        data = pd.read_csv('tests/data/arrivals.csv')
        group = 'Country'
        train = pd.DataFrame(columns=data.columns)
        test = pd.DataFrame(columns=data.columns)

        train_ratio = 0.8
        for g in data[group].unique():
            subframe = data[data[group] == g]
            length = subframe.shape[0]
            train = train.append(subframe[:int(length * train_ratio)])
            test = test.append(subframe[int(length * train_ratio):])

        target = 'Traffic'
        nr_preds = 2
        order_by = 'T'

        # Test multiple predictors playing along together
        a = {}
        a[1] = predictor_from_problem(pd.read_csv('tests/data/hdi.csv'), ProblemDefinition.from_dict(
            {'target': 'Development Index', 'time_aim': 10}))
        a[1].learn(pd.read_csv('tests/data/hdi.csv'))
        a[1].predict(pd.read_csv('tests/data/hdi.csv').iloc[0:10])

        prdb = ProblemDefinition.from_dict({'target': target,
                                            'time_aim': 30,
                                            'nfolds': 10,
                                            'anomaly_detection': True,
                                            'timeseries_settings': {
                                                'use_previous_target': True,
                                                'group_by': ['Country'],
                                                'nr_predictions': nr_preds,
                                                'order_by': [order_by],
                                                'window': 5
                                            }
                                            })

        a[2] = predictor_from_problem(train, prdb)

        a[2].learn(train)
        preds = a[2].predict(test)
        self.check_ts_prediction_df(preds, nr_preds, [order_by])

        # test inferring mode
        test['__mdb_make_predictions'] = False
        preds = a[2].predict(test)
        self.check_ts_prediction_df(preds, nr_preds, [order_by])

        # Additionally, check timestamps are further into the future than test dates
        latest_timestamp = pd.to_datetime(test[order_by]).max().timestamp()
        for idx, row in preds.iterrows():
            for timestamp in row[f'order_{order_by}']:
                assert timestamp > latest_timestamp

    def test_2_time_series_classification(self):
        from lightwood.api.high_level import predictor_from_problem

        df = pd.read_csv('tests/data/arrivals.csv')
        target = 'Traffic'
        df[target] = df[target] > 100000

        train_idxs = np.random.rand(len(df)) < 0.8
        train = df[train_idxs]
        test = df[~train_idxs]

        predictor = predictor_from_problem(df,
                                           ProblemDefinition.from_dict({'target': target,
                                                                        'time_aim': 30,
                                                                        'nfolds': 5,
                                                                        'anomaly_detection': False,
                                                                        'timeseries_settings': {
                                                                            'order_by': ['T'],
                                                                            'use_previous_target': True,
                                                                            'window': 5
                                                                        },
                                                                        }))

        predictor.learn(train)
        predictor.predict(test)
